from tensorflow.keras.layers import Dense, Conv2D, MaxPooling2D, Dropout, Input, Flatten, AveragePooling2D
from tensorflow.keras.models import Model


def model_rep(top, acti_size):
    input_ = Input(shape=(top, acti_size, 1))
    out = Conv2D(128, kernel_size=(2, 2), activation='relu')(input_)
    out = AveragePooling2D(pool_size=(2, 2))(out)
    out = Dropout(0.3)(out)
    out = Conv2D(64, kernel_size=(2, 2), activation='relu')(out)
    out = AveragePooling2D(pool_size=(2, 2))(out)
    out = Dropout(0.3)(out)
    out = Conv2D(32, kernel_size=(2, 2), activation='relu')(out)
    out = AveragePooling2D(pool_size=(2, 2))(out)
    out = Dropout(0.3)(out)
    out = Conv2D(16, kernel_size=(2, 2), activation='relu')(out)
    out = AveragePooling2D(pool_size=(2, 2))(out)
    out = Dropout(0.3)(out)
    out = Flatten()(out)
    out = Dense(units=1024, activation='relu')(out)
    out = Dense(units=1, activation='sigmoid')(out)

    return Model(input_, out)



import numpy as np
import os
from hipo_map.WSI_Preprocessing.Preprocessing import WSI_Scanning, Utilities
from tensorflow.keras.models import Model, load_model
import seaborn as sns
import matplotlib.pylab as plt

def ReLU(x):
    return x * (x > 0)


def extractingPatches(inputsvs, magnification, patch_size, Annotation=None,
                      Annotatedlevel=0, Requiredlevel=0, intermediate_layer_model=None):
    
    slide, _ = WSI_Scanning.readWSI(inputsvs, magnification, Annotation, Annotatedlevel, Requiredlevel)
    ALL = []
    for i in range(int(len(slide[0]) / 299)):
        for j in range(int(len(slide) / 599)):
            centerpoint = ((j * 299 + (patch_size[0] / 2)), (i * 299 + (patch_size[1] / 2)))
            sample_img = slide[int(centerpoint[0] - patch_size[0] / 2):int(centerpoint[0] + patch_size[0] / 2),
                         int(centerpoint[1] - patch_size[1] / 2): int(centerpoint[1] + patch_size[1] / 2)]
            patchs = Utilities.stainremover_small_patch_remover1(sample_img, patch_size)
            if patchs is None:
                None
            else:
                patchs = patchs / 255
                patchs = np.expand_dims(patchs, axis=0)
                intermediate_output, y_pred = intermediate_layer_model.predict(patchs)
                _, width, height, _ = intermediate_output.shape
                y = y_pred[0][0]
                al_patch = np.zeros((1, width, height))
                for t in range(len(intermediate_output[0][0][0])):
                    grad = np.gradient(intermediate_output[:, :, :, t].flatten(), abs(1 - y + 0.0001)) / 64
                    al = sum(grad)
                    al_patch += al * intermediate_output[:, :, :, t]
                al_patch = ReLU(al_patch)
                ALL.append(sorted(al_patch.flatten(), reverse=True))
    ALL = np.sort(np.array(ALL), axis=0)[::-1]
    return ALL


def generateHipoMap(inputpath, outputpath, magnification="20x", patch_size=(299, 299),
                model=None, layer_name=None, Annotation=None, Annotatedlevel=0, Requiredlevel=0):
        
    if model == None:
        print("Error : Model not provided as parameter")
    if layer_name == None:
        print("Error : layer_name not provided as parameter")
    else:
        intermediate_layer_model = Model(inputs=model.input, outputs=[model.get_layer(layer_name).output, model.output])
        list_wsi = os.listdir(inputpath)
        for wsi in list_wsi:
            if wsi[-3:] == 'svs':
                print("start " + wsi)
                data = inputpath + wsi
                print(data)
                repmap = extractingPatches(data, magnification, patch_size, Annotation, Annotatedlevel, Requiredlevel,intermediate_layer_model)
                print("done")
                save_path = outputpath + wsi[:-4]
                np.save(save_path, repmap)
                
def draw_represent(path, K, max_value=1000):
    list_represent = os.listdir(path)
    for rep in list_represent:
        print(path + rep)
        try:
            plt.clf()
            heat = np.load(path + rep)
            plt.title("Heatmap for " + rep[:-3], fontsize=20)
            plt.xlabel("pixel of activation map", fontsize=5)
            plt.ylabel("patches", fontsize=5)
            ax = sns.heatmap(heat[:K], vmax=max_value)
            plt.show()
            plt.pause(0.1)
        except:
            pass
    


from hipo_map.model_rep import model_rep
import numpy as np
from tensorflow.keras import optimizers
import tensorflow as tf
from scipy import interp
from sklearn import metrics
import pandas as pd
import os
import seaborn as sns
import matplotlib.pylab as plt

        
class HipoMap:
    
    def __init__(self, K):
        self.K = K

    def fit(self, train_set, valid_set, lr, epoch, batchsize, activation_size):
        self.activation_size = activation_size
        
        train_X, train_y = train_set
        valid_X, valid_y = valid_set

        self.mean = np.mean(train_X, axis=0)
        self.std = np.std(train_X, axis=0)

        train_X = (train_X - self.mean) / self.std
        train_X = np.reshape(train_X, (len(train_X), self.K, self.activation_size))
        train_X = np.expand_dims(train_X, axis=3)
        train_y = np.array(train_y)

        valid_X = (valid_X - self.mean) / self.std
        valid_X = np.reshape(valid_X, (len(valid_X), self.K, self.activation_size))
        valid_X = np.expand_dims(valid_X, axis=3)
        valid_y = np.array(valid_y)

        SGD = optimizers.SGD(lr=lr)
        model = model_rep(self.K, self.activation_size)

        model.compile(optimizer=SGD, loss = "binary_crossentropy", metrics=['accuracy', tf.keras.metrics.AUC()])
        model.fit(train_X, train_y, batch_size=batchsize, epochs=epoch, validation_data=(valid_X, valid_y))

        return model

    def predict(self, model, test_X):
        test_X = (test_X - self.mean) / self.std
        test_X = np.reshape(test_X, (len(test_X), self.K, self.activation_size))
        test_X = np.expand_dims(test_X, axis=3)

        prediction = model.predict(test_X)

        return prediction

    def evaluate_score(self, label, prediction):
        label = np.array(label)
        fpr, tpr, threshold = metrics.roc_curve(label, prediction)
        mean_fpr = np.linspace(0, 1, 100)
        tpr_score = interp(mean_fpr, fpr, tpr)
        auc_score = metrics.auc(fpr,tpr)

        return tpr_score, mean_fpr, auc_score
    
    

    def split(self, split_csv, dir_normal, dir_cancer):
        splits = pd.read_csv(split_csv, header=None)
        trains = splits[splits[3] == 'train']
        valids = splits[splits[3] == 'valid']
        tests = splits[splits[3] == 'test']

        X_train, y_train = self.load_Dataset(trains, dir_normal, dir_cancer)
        X_valid, y_valid = self.load_Dataset(valids, dir_normal, dir_cancer)
        X_test, y_test = self.load_Dataset(tests, dir_normal, dir_cancer)

        return (X_train, y_train), (X_valid, y_valid), (X_test, y_test)

    def load_Dataset(self, base, dir_normal, dir_cancer):
        list_X, list_y = [], []
        normal_set = os.listdir(dir_normal)
        cancer_set = os.listdir(dir_cancer)

        for data, label in zip(base[1], base[2]):
            if label == 'normal':
                for normal in normal_set:
                    if data in normal:
                        img = np.load(dir_normal + normal)
                        if img.shape[0] >= self.K:
                            list_X.append(img[:self.K].flatten())
                            list_y.append(0)
                        break
            else:
                for cancer in cancer_set:
                    if data in cancer:
                        img = np.load(dir_cancer + cancer)
                        if img.shape[0] >= self.K:
                            list_X.append(img[:self.K].flatten())
                            list_y.append(1)
                        break

        return np.array(list_X), np.array(list_y)



    
    
